"""Automower library using aiohttp."""
from __future__ import annotations

import logging
import time
from urllib.parse import quote_plus, urlencode

import aiohttp

from .const import (
    AUTH_API_REVOKE_URL,
    AUTH_API_TOKEN_URL,
    AUTH_HEADERS,
    MOWER_API_BASE_URL,
    TOKEN_URL,
)

_LOGGER = logging.getLogger(__name__)

timeout = aiohttp.ClientTimeout(total=10)


class TokenError(Exception):
    """Raised when Husqvarna Authentication API request ended in error 400."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status


class TokenRefreshError(Exception):
    """Raised when Husqvarna Authentication API is not able to refresh the token (Error 400 or 404)."""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status


class TokenValidationError(Exception):
    """Raised when Husqvarna Authentication API token request ended in error 404. The reason might be an invalid token or that a refresh is needed"""

    def __init__(self, status: str) -> None:
        """Initialize."""
        super().__init__(status)
        self.status = status


class GetAccessTokenClientCredentials:
    """Class to get an acces token from the Authentication API with client_credentials.
    This grant type is intended only for you. If you want other users to use your application,
    then they should login using Authorization Code Grant."""

    def __init__(self, client_id, client_secret) -> None:
        """Initialize the Auth-API and store the auth so we can make requests."""
        self.auth_data = urlencode(
            {
                "grant_type": "client_credentials",
                "client_id": client_id,
                "client_secret": client_secret,
            },
            quote_via=quote_plus,
        )

    async def async_get_access_token(self) -> dict:
        """Return the token."""
        async with aiohttp.ClientSession(headers=AUTH_HEADERS) as session:
            async with session.post(AUTH_API_TOKEN_URL, data=self.auth_data) as resp:
                result = await resp.json(encoding="UTF-8")
                _LOGGER.debug("Resp.status get access token: %s", result)
                if resp.status == 200:
                    result = await resp.json(encoding="UTF-8")
                    result["expires_at"] = result["expires_in"] + time.time()
                if resp.status >= 400:
                    raise TokenError(
                        f"The token is invalid, respone from Husqvarna Automower API: {result}"
                    )
        result["status"] = resp.status
        return result


class RefreshAccessToken:
    """Class to renew the Access Token."""

    def __init__(self, api_key, refresh_token) -> None:
        """Initialize the Auth-API and store the auth so we can make requests."""
        self.api_key = api_key
        self.refresh_token = refresh_token
        self.auth_data = urlencode(
            {
                "client_id": self.api_key,
                "grant_type": "refresh_token",
                "refresh_token": self.refresh_token,
            },
            quote_via=quote_plus,
        )

    async def async_refresh_access_token(self) -> dict:
        """Return the refresh token."""
        async with aiohttp.ClientSession(headers=AUTH_HEADERS) as session:
            async with session.post(AUTH_API_TOKEN_URL, data=self.auth_data) as resp:
                result = await resp.json(encoding="UTF-8")
                _LOGGER.debug("Resp.status refresh token: %s", result)
                if resp.status == 200:
                    result["expires_at"] = result["expires_in"] + time.time()
                    result["status"] = resp.status
                    return result
                if resp.status in [400, 401, 404]:
                    raise TokenRefreshError(
                        f"The token cannot be refreshed, respone from Husqvarna Automower API: {result}"
                    )


class RevokeAccessToken:
    """Class to invalidate an access token."""

    def __init__(self, access_token) -> None:
        """Initialize the Auth-API and store the auth so we can make requests."""
        self.access_token = access_token
        self.auth_data = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Bearer {0}".format(self.access_token),
            "Accept": "*/*",
        }

    async def async_delete_access_token(self) -> dict:
        """Delete the token."""
        async with aiohttp.ClientSession(headers=self.auth_data) as session:
            async with session.post(
                AUTH_API_REVOKE_URL, data=(f"token={self.access_token}")
            ) as resp:
                result = await resp.json(encoding="UTF-8")
                _LOGGER.debug("Resp.status delete token: %s", resp.status)
                if resp.status >= 400:
                    resp.raise_for_status()
                    _LOGGER.error("Response body delete token: %s", result)
        return result


class GetMowerData:
    """Class to communicate with the Automower Connect API."""

    def __init__(self, api_key, access_token, provider, token_type) -> None:
        """Initialize the Communication API to get data."""
        self.api_key = api_key
        self.access_token = access_token
        self.provider = provider
        self.token_type = token_type
        self.mower_headers = {
            "Authorization": "{0} {1}".format(self.token_type, self.access_token),
            "Authorization-Provider": "{0}".format(self.provider),
            "Content-Type": "application/vnd.api+json",
            "X-Api-Key": "{0}".format(self.api_key),
        }

    async def async_mower_state(self) -> list[dict]:
        """Return the mowers data as a list of mowers."""
        async with aiohttp.ClientSession(
            headers=self.mower_headers, timeout=timeout
        ) as session:
            async with session.get(MOWER_API_BASE_URL) as resp:
                result = await resp.json(encoding="UTF-8")
                _LOGGER.debug("Response mower data: %s", resp)
                if resp.status == 200:
                    result = await resp.json(encoding="UTF-8")
                    for idx, ent in enumerate(result["data"]):
                        result["data"][idx]["attributes"].update(
                            result["data"][idx]["attributes"]["settings"]
                        )
                        del result["data"][idx]["attributes"]["settings"]
                    _LOGGER.debug("Result mower data: %s", result)
                if resp.status >= 400:
                    _LOGGER.error("Response mower data: %s", result)
        return result


class Return:
    """Class to send commands to the Automower Connect API."""

    def __init__(
        self,
        api_key,
        access_token,
        provider,
        token_type,
        mower_id,
        payload,
        command_type,
    ) -> None:
        """Initialize the API and store the auth so we can send commands."""
        self.api_key = api_key
        self.access_token = access_token
        self.provider = provider
        self.token_type = token_type
        self.mower_id = mower_id
        self.command_type = command_type
        self.mower_headers = {
            "Authorization": "{0} {1}".format(self.token_type, self.access_token),
            "Authorization-Provider": "{0}".format(self.provider),
            "Content-Type": "application/vnd.api+json",
            "accept": "*/*",
            "X-Api-Key": "{0}".format(self.api_key),
        }
        self.mower_action_url = (
            f"{MOWER_API_BASE_URL}{self.mower_id}/{self.command_type}"
        )
        self.payload = payload

    async def async_mower_command(self) -> None:
        """Send a payload to the mower to execute a command."""
        async with aiohttp.ClientSession(headers=self.mower_headers) as session:
            async with session.post(self.mower_action_url, data=self.payload) as resp:
                await session.close()
        _LOGGER.debug("Mower Action URL: %s", self.mower_action_url)
        _LOGGER.debug("Sent payload: %s", self.payload)
        _LOGGER.debug("Resp status mower command: %s", resp.status)
        if resp.status >= 400:
            resp.raise_for_status()



"""Module to connect to Automower with websocket."""
import asyncio
import datetime
import logging
import time

import aiohttp

from . import rest
from .const import (
    AUTH_HEADER_FMT,
    EVENT_TYPES,
    HUSQVARNA_URL,
    MARGIN_TIME,
    MIN_SLEEP_TIME,
    REST_POLL_CYCLE,
    REST_POLL_CYCLE_LE,
    WS_STATUS_UPDATE_CYLE,
    WS_TOLERANCE_TIME,
    WS_URL,
)

_LOGGER = logging.getLogger(__name__)


class AutomowerSession:
    """Session"""

    def __init__(
        self,
        api_key: str,
        token: dict = None,
        low_energy=True,
        ws_heartbeat_interval: float = 60.0,
        loop=None,
    ):
        """Create a session.

        :param str api_key: A 36 digit api key.
        :param dict token: A token as returned by rest.GetAccessToken.async_get_access_token()
        :param float ws_heartbeat_interval: Periodicity of keep-alive pings on the websocket in seconds.
        :param loop: Event-loop for task execution. If None, the event loop in the current OS thread is used.
        """
        self.api_key = api_key
        self.token = token
        self.data_update_cbs = []
        self.token_update_cbs = []
        self.ws_heartbeat_interval = ws_heartbeat_interval
        self.rest_task = False
        self.low_energy = low_energy
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.data = None

        self.ws_task = None
        self.token_task = None
        self.websocket_monitor_task = None
        self.rest_task = None

    def register_data_callback(self, callback, schedule_immediately=False):
        """Register a data update callback.

        :param func callback: Callback fired on data updates. Takes one dict argument which is the up-to-date mower data list.
        :param bool schedule_immediately: Schedule callback immediately (if data is available).
        """
        if callback not in self.data_update_cbs:
            self.data_update_cbs.append(callback)
        if schedule_immediately:
            self._schedule_data_callback(
                callback, delay=1e-3
            )  # Need a delay for home assistant to finish entity setup.

    def unregister_data_callback(self, callback):
        """Unregister a data update callback.

        :param func callback: Callback fired on data updates. Takes one dict argument which is the up-to-date mower data list.
        """
        if callback in self.data_update_cbs:
            self.data_update_cbs.remove(callback)

    def register_token_callback(self, callback, schedule_immediately=False):
        """Register a token update callback.

        :param func callback: Callback fired on token updates. Takes one dict argument which is the newly received token.
        :param bool schedule_immediately: Schedule callback immediately (if token is available).
        """
        if callback not in self.token_update_cbs:
            self.token_update_cbs.append(callback)
        if schedule_immediately:
            self._schedule_token_callback(
                callback, delay=1e-3
            )  # Need a delay for home assistant to finish entity setup.

    async def logincc(self, client_secret: str) -> dict:
        """Login with client credentials.

        This method gets an access token with a client_id (Api key) and a client_secret.
        This token can't be refreshed. Create a new one after it is expired.

        :param str client_secret: Your client_secret
        :return dict: The token as returned by
        rest.GetAccessTokenClientCredentials.async_get_access_token().
        You can store this persistently and pass it to the constructor
        on subsequent instantiations.
        """
        a = rest.GetAccessTokenClientCredentials(self.api_key, client_secret)
        self.token = await a.async_get_access_token()
        self._schedule_token_callbacks()
        return self.token

    async def connect(self):
        """Connect to the API.

        This method handles the login and starts a task that keep the access
        token constantly fresh. This method works only, if the token is created with the
        Authorization Code Grant. Call this method before any other methods.
        """
        if self.token is None:
            raise AttributeError("No token to connect with.")
        if time.time() > (self.token["expires_at"] - MARGIN_TIME):
            await self.refresh_token()

        self.data = await self.get_status()
        self._schedule_data_callbacks()

        if "amc:api" not in self.token["scope"]:
            _LOGGER.error(
                "Your API-Key is not compatible to the websocket, please refresh it on %s",
                HUSQVARNA_URL,
            )
        else:
            self.ws_task = self.loop.create_task(self._ws_task())
        self.rest_task = self.loop.create_task(self._rest_task())
        self.token_task = self.loop.create_task(self._token_monitor_task())

    async def close(self):
        """Close the session."""
        await self.invalidate_token()
        for task in [
            self.ws_task,
            self.token_task,
            self.websocket_monitor_task,
            self.rest_task,
        ]:
            tasks = []
            if task is not None:
                tasks.append(task)
                if not task.cancelled():
                    task.cancel()
        try:
            await asyncio.gather(*tasks)
        except asyncio.CancelledError:
            pass

    async def get_status(self):
        """Get mower status via Rest."""
        if self.token is None:
            _LOGGER.warning("No token available")
            return None
        d = rest.GetMowerData(
            self.api_key,
            self.token["access_token"],
            self.token["provider"],
            self.token["token_type"],
        )
        return await d.async_mower_state()

    async def action(self, mower_id, payload, command_type):
        """Send command to the mower via Rest."""
        if self.token is None:
            _LOGGER.warning("No token available")
            return None
        a = rest.Return(
            self.api_key,
            self.token["access_token"],
            self.token["provider"],
            self.token["token_type"],
            mower_id,
            payload,
            command_type,
        )
        return await a.async_mower_command()

    async def invalidate_token(self):
        """Invalidate token via Rest."""
        if self.token is None:
            _LOGGER.warning("No token available")
            return None
        token = rest.RevokeAccessToken(self.token["access_token"])
        return await token.async_delete_access_token()

    async def refresh_token(self):
        """Refresh token via Rest."""
        if "refresh_token" not in self.token:
            _LOGGER.warning("No refresh token available")
            return None
        _LOGGER.debug("Refresh access token")
        r = rest.RefreshAccessToken(self.api_key, self.token["refresh_token"])
        self.token = await r.async_refresh_access_token()
        self._schedule_token_callbacks()

    async def _token_monitor_task(self):
        while True:
            if "expires_at" in self.token:
                expires_at = self.token["expires_at"]
                sleep_time = max(MIN_SLEEP_TIME, expires_at - time.time() - MARGIN_TIME)
            else:
                sleep_time = MIN_SLEEP_TIME

            _LOGGER.debug("token_monitor_task sleeping for %s sec", sleep_time)
            await asyncio.sleep(sleep_time)
            await self.refresh_token()

    def _update_data(self, j):
        if self.data is None:
            _LOGGER.error("Failed to update data with ws response (no data)")
            return
        for datum in self.data["data"]:
            if datum["type"] == "mower" and datum["id"] == j["id"]:
                for attrib in j["attributes"]:
                    try:
                        tasks = j["attributes"]["calendar"]["tasks"]
                        if len(tasks) == 0:
                            temp_task = datum["attributes"]["calendar"]["tasks"]
                            datum["attributes"][attrib] = j["attributes"][attrib]
                            datum["attributes"]["calendar"]["tasks"] = temp_task
                        if len(tasks) > 0:
                            datum["attributes"][attrib] = j["attributes"][attrib]
                    except KeyError:
                        datum["attributes"][attrib] = j["attributes"][attrib]
                return
        _LOGGER.error("Failed to update data with ws response (id not found)")

    def _schedule_token_callback(self, cb, delay=0.0):
        if self.token is None:
            _LOGGER.debug("No token available. Will not schedule callback")
            return
        self.loop.call_later(delay, cb, self.token)

    def _schedule_token_callbacks(self):
        for cb in self.token_update_cbs:
            self._schedule_token_callback(cb)

    def _schedule_data_callback(self, cb, delay=0.0):
        if self.data is None:
            _LOGGER.debug("No data available. Will not schedule callback")
            return
        self.loop.call_later(delay, cb, self.data)

    def _schedule_data_callbacks(self):
        for cb in self.data_update_cbs:
            self._schedule_data_callback(cb)

    async def _ws_task(self):
        printed_err_msg = False
        self.websocket_monitor_task = self.loop.create_task(
            self._websocket_monitor_task()
        )
        async with aiohttp.ClientSession() as session:
            while True:
                if self.token is None or "access_token" not in self.token:
                    if not printed_err_msg:
                        # New login() needed but since we don't store username
                        # and password, we cannot get request one.
                        #
                        # TODO: Add callback for this to notify the user that
                        # this has happened.
                        _LOGGER.warning("No access token for ws auth. Retrying")
                        printed_err_msg = True
                    await asyncio.sleep(60.0)
                    continue
                printed_err_msg = False
                async with session.ws_connect(
                    url=WS_URL,
                    headers={
                        "Authorization": AUTH_HEADER_FMT.format(
                            self.token["access_token"]
                        )
                    },
                    heartbeat=self.ws_heartbeat_interval,
                ) as ws:
                    _LOGGER.debug("Websocket (re)connected")
                    async for msg in ws:
                        if msg.type == aiohttp.WSMsgType.TEXT:
                            j = msg.json()
                            if "type" in j:
                                if j["type"] in EVENT_TYPES:
                                    self._update_data(j)
                                    _LOGGER.debug("Got %s, data: %s", j["type"], j)
                                    self._schedule_data_callbacks()
                                else:
                                    _LOGGER.warning(
                                        "Received unknown ws type %s", j["type"]
                                    )
                            elif "ready" in j and "connectionId" in j:
                                _LOGGER.debug(
                                    "Websocket ready=%s (id='%s')",
                                    j["ready"],
                                    j["connectionId"],
                                )
                            else:
                                _LOGGER.debug("Discarded websocket response: %s", j)
                        elif msg.type == aiohttp.WSMsgType.ERROR:
                            _LOGGER.debug("Received ERROR")
                            break
                        elif msg.type == aiohttp.WSMsgType.CONTINUATION:
                            _LOGGER.debug("Received CONTINUATION")
                        elif msg.type == aiohttp.WSMsgType.BINARY:
                            _LOGGER.debug("Received BINARY")
                        elif msg.type == aiohttp.WSMsgType.PING:
                            _LOGGER.debug("Received PING")
                        elif msg.type == aiohttp.WSMsgType.PONG:
                            _LOGGER.debug("Received PONG")
                        elif msg.type == aiohttp.WSMsgType.CLOSE:
                            _LOGGER.debug("Received CLOSE")
                        elif msg.type == aiohttp.WSMsgType.CLOSING:
                            _LOGGER.debug("Received CLOSING")
                        elif msg.type == aiohttp.WSMsgType.CLOSED:
                            _LOGGER.debug("Received CLOSED")
                        else:
                            _LOGGER.debug("Received msg.type=%d", msg.type)

    async def _rest_task(self):
        """Poll data periodically via Rest."""
        while True:
            _LOGGER.debug(
                "LE: %s",
                self.low_energy,
            )
            if self.low_energy is True:
                await asyncio.sleep(REST_POLL_CYCLE_LE)
            if self.low_energy is False:
                await asyncio.sleep(REST_POLL_CYCLE)
            self.data = await self.get_status()
            self._schedule_data_callbacks()

    async def _websocket_monitor_task(self):
        """Monitor, if the websocket still sends updates. If not, check, via REST,
        if the mower is connected. If there are no recent updates, Start REST task
        to get information"""
        message_sent = []
        mower_connected = []
        for idx, ent in enumerate(self.data["data"]):
            mower_connected.append(
                self.data["data"][idx]["attributes"]["metadata"]["connected"]
            )
            message_sent.append(not mower_connected[idx])
        while True:
            for idx, ent in enumerate(self.data["data"]):
                mower_connected[idx] = self.data["data"][idx]["attributes"]["metadata"][
                    "connected"
                ]
                if not mower_connected[idx] and not message_sent[idx]:
                    message_sent[idx] = True
                    _LOGGER.warning(
                        "Connection to %s lost",
                        self.data["data"][idx]["attributes"]["system"]["name"],
                    )
                if mower_connected[idx] and message_sent[idx]:
                    message_sent[idx] = False
                    _LOGGER.info(
                        "Connected to %s again",
                        self.data["data"][idx]["attributes"]["system"]["name"],
                    )
                timestamp = (
                    self.data["data"][idx]["attributes"]["metadata"]["statusTimestamp"]
                    / 1000
                )
                now = datetime.datetime.now().timestamp()
                age = now - timestamp
                _LOGGER.debug("Age in sec: %i", age)
            ws_monitor_sleep_time = min(
                max(WS_STATUS_UPDATE_CYLE + WS_TOLERANCE_TIME - age, REST_POLL_CYCLE),
                WS_STATUS_UPDATE_CYLE + WS_TOLERANCE_TIME,
            )
            if age < (WS_STATUS_UPDATE_CYLE + WS_TOLERANCE_TIME):
                _LOGGER.debug(
                    "websocket_monitor_task sleeping for %ss", ws_monitor_sleep_time
                )
            any_mowers_connected = any(mower_connected)
            if age > (WS_STATUS_UPDATE_CYLE + WS_TOLERANCE_TIME):
                if not any_mowers_connected:
                    _LOGGER.debug("No ws updates anymore, and mower disconnected")
                if any_mowers_connected:
                    _LOGGER.debug(
                        "No ws updates anymore and mower connected, ws probably down or mower shortly before disconnecting"
                    )
            await asyncio.sleep(ws_monitor_sleep_time)



"""The constants for aioautomower."""
AUTH_API_BASE_URL = "https://api.authentication.husqvarnagroup.dev/v1"
AUTH_API_TOKEN_URL = f"{AUTH_API_BASE_URL}/oauth2/token"
AUTH_API_REVOKE_URL = f"{AUTH_API_BASE_URL}/oauth2/revoke"
AUTH_HEADERS = {
    "Content-Type": "application/x-www-form-urlencoded",
    "Accept": "application/json",
}
AUTH_HEADER_FMT = "Bearer {}"
EVENT_TYPES = [
    "status-event",
    "positions-event",
    "settings-event",
]
HUSQVARNA_URL = "https://developer.husqvarnagroup.cloud/"
MARGIN_TIME = 60.0  # Token is typically valid for 24h, request a new one some time before its expiration to avoid glitches.
MIN_SLEEP_TIME = 600.0  # Avoid hammering
MOWER_API_BASE_URL = "https://api.amc.husqvarna.dev/v1/mowers/"
REST_POLL_CYCLE = 300
REST_POLL_CYCLE_LE = 86400
TOKEN_URL = f"{AUTH_API_BASE_URL}/token"
USER_URL = f"{AUTH_API_BASE_URL}/users"
WS_STATUS_UPDATE_CYLE = 840.0
WS_TOLERANCE_TIME = 20.0
WS_URL = "wss://ws.openapi.husqvarna.dev/v1"



#!/usr/bin/env python3
import asyncio
import json
import logging
import signal

import aioautomower

_LOGGER = logging.getLogger(__name__)


async def run_tester(client_secret: str, api_key: str, token: dict):
    sess = aioautomower.AutomowerSession(api_key, token, ws_heartbeat_interval=60)
    if sess.token is None:
        token = await sess.logincc(client_secret)

    sess.register_data_callback(
        lambda x: logging.info("data callback;%s" % json.dumps(x)),
        schedule_immediately=False,
    )
    sess.register_token_callback(
        lambda x: logging.info("token callback;%s" % json.dumps(x)),
        schedule_immediately=True,
    )

    await sess.connect()

    def sigusr1():
        asyncio.ensure_future(sess.invalidate_token())

    def sigusr2():
        asyncio.ensure_future(sess.get_status())

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGUSR1, sigusr1)
    loop.add_signal_handler(signal.SIGUSR2, sigusr2)

    while True:
        await asyncio.sleep(0.1)


def main():
    """Tester for the Husqvarna Automower API

    The tester will login using username and password and connect to a
    websocket listening for mower updates.

    The tester listens to two signals on which it performs the following
    requests:
    SIGUSR1: Invalidate token
    SIGUSR2: Get status"""
    import argparse

    parser = argparse.ArgumentParser(
        description=main.__doc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "-s", "--client_secret", required=True, help="Husqvarna app username"
    )
    parser.add_argument("-k", "--api-key", required=True, help="Husqvarna API key")
    parser.add_argument(
        "-t",
        "--token",
        nargs="?",
        type=argparse.FileType("r"),
        default=None,
        help="Optional access token. If provided, username and password will not be used.",
    )

    args = parser.parse_args()

    token = json.load(args.token) if args.token is not None else None

    logging.basicConfig(level="DEBUG", format="%(asctime)s;%(levelname)s;%(message)s")
    asyncio.run(run_tester(args.client_secret, args.api_key, token))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aioautomower",
    author="Thomas Protzner",
    author_email="thomas.protzner@gmail.com",
    description="module to communicate to Husqvarna Automower API",
    license="Apache License 2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Thomas55555/aioautomower",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=list(val.strip() for val in open("requirements.txt")),
    version="2023.4.2",
    entry_points={
        "console_scripts": ["automower=aioautomower.cli:main"],
    },
)
